@@ -1,6 +1,7 @@
-/* Copyright (c) 1997-2018
-   Ewgenij Gawrilow, Michael Joswig (Technische Universitaet Berlin, Germany)
-   http://www.polymake.org
+/* Copyright (c) 1997-2023
+   Ewgenij Gawrilow, Michael Joswig, and the polymake team
+   Technische Universität Berlin, Germany
+   https://polymake.org
 
    This program is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
@@ -16,17 +17,12 @@
 
 #include "polymake/perl/Ext.h"
 
-namespace pm { namespace perl {
-namespace glue {
-
-Perl_ppaddr_t def_pp_LEAVE;
-
-}
-
-using namespace pm::perl::glue;
+namespace pm { namespace perl { namespace glue {
 
 namespace {
 
+Perl_ppaddr_t def_pp_LEAVE, def_pp_OPEN;
+
 template <typename LocalHandler>
 struct local_wrapper {
    static void undo(pTHX_ void* p)
@@ -86,14 +82,35 @@ struct local_ref_handler {
    }
 };
 
+}
+
+OP* parse_expression_in_parens(pTHX)
+{
+   lex_read_space(0);
+   if (PL_parser->bufptr == PL_parser->bufend || *PL_parser->bufptr != '(')
+      return nullptr;
+   lex_read_to(PL_parser->bufptr+1);
+   OP* o = parse_termexpr(0);
+   if (!o) return nullptr;
+   lex_read_space(0);
+   if (PL_parser->bufptr == PL_parser->bufend || *PL_parser->bufptr != ')') {
+      op_free(o);
+      return nullptr;
+   }
+   lex_read_to(PL_parser->bufptr+1);
+   return o;
+}
+
 }
 namespace ops {
 
+using namespace pm::perl::glue;
+
 OP* local_ref(pTHX)
 {
    dSP;
-   SV *left = POPs;
-   SV *right = GIMME_V != G_VOID ? TOPs : POPs;
+   SV* left = POPs;
+   SV* right = GIMME_V != G_VOID ? TOPs : POPs;
    if (!SvROK(right))
       DIE(aTHX_ "local ref value must be a reference");
    SV* value = SvRV(right);
@@ -140,7 +157,7 @@ OP* local_ref(pTHX)
 }
 
 }
-namespace {
+namespace glue { namespace {
 
 int parse_local_ref(pTHX_ OP** op_ptr)
 {
@@ -198,17 +215,23 @@ OP* local_scalar_op(pTHX)
 
 // --------------------
 
-}
+} }
 namespace ops {
 
+void
+localize_scalar(pTHX_ SV* var, SV* value)
+{
+   local_do<local_scalar_handler>(aTHX_ var, value);
+}
+
 void
 localize_scalar(pTHX_ SV* var)
 {
-   local_do<local_scalar_handler>(aTHX_ var, sv_mortalcopy(var));
+   localize_scalar(aTHX_ var, sv_mortalcopy(var));
 }
 
 }
-namespace {
+namespace glue { namespace {
 
 OP* local_save_scalar_op(pTHX)
 {
@@ -233,7 +256,7 @@ struct local_incr_handler {
       if (SvIOK(var))
          sv_setiv(var, SvIVX(var) - incr);
       else if (SvNOK(var))
-         sv_setnv(var, SvNVX(var) - incr);
+         sv_setnv(var, SvNVX(var) - NV(incr));
       else
          Perl_croak(aTHX_ "undoing local increment: variable is no more numerical");
    }
@@ -321,12 +344,12 @@ struct local_push_handler : local_push_unshift_handler {
       : local_push_unshift_handler(av_, n_)
    {
       av_extend(av, AvFILLp(av) + n);
-      insert_elems(aTHX_ src, AvARRAY(av) + AvFILLp(av) + 1);
+      insert_elems(aTHX_ src, AvARRAY(av)+AvFILLp(av)+1);
    }
 
    void undo(pTHX) const
    {
-      for (SV **e = AvARRAY(av) + AvFILLp(av), **stop = e - n; e > stop; --e) {
+      for (SV **e = AvARRAY(av) + AvFILLp(av), **stop = e-n; e > stop; --e) {
          SvREFCNT_dec(*e);
          *e = PmEmptyArraySlot;
       }
@@ -339,21 +362,21 @@ struct local_unshift_handler : local_push_unshift_handler {
    local_unshift_handler(pTHX_ AV* av_, SV* const * src, IV n_)
       : local_push_unshift_handler(av_, n_)
    {
-      av_extend(av, AvFILLp(av) + n);
+      av_extend(av, AvFILLp(av)+n);
       SV** dst = AvARRAY(av);
-      Move(dst, dst + n_, AvFILLp(av) + 1, SV*);
+      Move(dst, dst + n_, AvFILLp(av)+1, SV*);
       insert_elems(aTHX_ src, dst);
    }
 
    void undo(pTHX) const
    {
       SV **e, **stop;
-      for (stop = AvARRAY(av) - 1, e = stop + n; e > stop; --e)
+      for (stop = AvARRAY(av)-1, e = stop + n; e > stop; --e)
          SvREFCNT_dec(*e);
       AvFILLp(av) -= n;
       ++stop;
-      Move(stop + n, stop, AvFILLp(av) + 1, SV*);
-      for (e = stop + AvFILLp(av) + 1, stop = e + n; e < stop; ++e)
+      Move(stop + n, stop, AvFILLp(av)+1, SV*);
+      for (e = stop + AvFILLp(av)+1, stop = e+n; e < stop; ++e)
          *e = PmEmptyArraySlot;
    }
 };
@@ -365,7 +388,7 @@ OP* local_push_unshift_op(pTHX)
    AV* av = (AV*)*++MARK;
    IV n = SP - MARK;
    if (n > 0)
-      local_do<std::conditional_t<is_unshift, local_unshift_handler, local_push_handler>>(aTHX_ av, MARK + 1, n);
+      local_do<std::conditional_t<is_unshift, local_unshift_handler, local_push_handler>>(aTHX_ av, MARK+1, n);
    SP = ORIGMARK;
    RETURN;
 }
@@ -388,15 +411,17 @@ struct local_pop_handler {
    SV* val;
 
    local_pop_handler(pTHX_ AV* av_)
-      : av(av_)
+      : av((AV*)SvREFCNT_inc_simple_NN(av_))
    {
-      SvREFCNT_inc_simple_void_NN(av);
       val = av_pop(av);
    }
 
    void undo(pTHX) const
    {
+      auto localizing = PL_localizing;
+      PL_localizing = 2;
       av_push(av, val);
+      PL_localizing = localizing;
       SvREFCNT_dec(av);
    }
 };
@@ -422,6 +447,7 @@ int parse_local_pop(pTHX_ OP** op_ptr)
    OP* o = parse_termexpr(0);
    if (!o) return KEYWORD_PLUGIN_DECLINE;
    o->op_ppaddr = local_pop_op;
+   o->op_private |= OPpLVAL_INTRO;
    *op_ptr = o;
    PL_hints |= HINT_BLOCK_SCOPE;
    return KEYWORD_PLUGIN_EXPR;
@@ -442,7 +468,10 @@ struct local_shift_handler {
 
    void undo(pTHX) const
    {
+      auto localizing = PL_localizing;
+      PL_localizing = 2;
       av_unshift(av, 1);
+      PL_localizing = localizing;
       AvARRAY(av)[0] = val;
       SvREFCNT_dec(av);
    }
@@ -469,6 +498,7 @@ int parse_local_shift(pTHX_ OP** op_ptr)
    OP* o = parse_termexpr(0);
    if (!o) return KEYWORD_PLUGIN_DECLINE;
    o->op_ppaddr = local_shift_op;
+   o->op_private |= OPpLVAL_INTRO;
    *op_ptr = o;
    PL_hints |= HINT_BLOCK_SCOPE;
    return KEYWORD_PLUGIN_EXPR;
@@ -509,9 +539,9 @@ OP* local_splice_op(pTHX)
 {
    dSP;  dMARK;  dORIGMARK;
    AV* av = (AV*)*++MARK;
-   if (MARK + 2 < SP)
+   if (MARK+2 < SP)
       DIE(aTHX_ "unsupported local splice with insertion");
-   const IV len = AvFILLp(av) + 1;
+   const IV len = AvFILLp(av)+1;
    IV first, size;
    if (MARK < SP) {
       ++MARK;
@@ -547,7 +577,7 @@ OP* local_splice_op(pTHX)
    if (size != 0) {
       if (GIMME_V == G_ARRAY) {
          EXTEND(SP, size);
-         Copy(AvARRAY(av) + first, SP + 1, size, SV*);
+         Copy(AvARRAY(av) + first, SP+1, size, SV*);
          SP += size;
       }
       local_do<local_splice_handler>(aTHX_ av, first, size);
@@ -593,8 +623,8 @@ OP* local_swap_op(pTHX)
    IV ix2 = POPi;
    IV ix1 = POPi;
    AV* av = (AV*)POPs;
-   if (ix1 < 0) ix1 += AvFILL(av) + 1;
-   if (ix2 < 0) ix2 += AvFILL(av) + 1;
+   if (ix1 < 0) ix1 += AvFILL(av)+1;
+   if (ix2 < 0) ix2 += AvFILL(av)+1;
    if (ix1 > AvFILL(av) || ix2 > AvFILL(av)) DIE(aTHX_ "local swap: index out of range");
    local_do<local_swap_handler>(aTHX_ av, ix1, ix2);
    RETURN;
@@ -700,7 +730,7 @@ bool save_localizations(pTHX_ I32& start, I32& end)
    I32 tmp_save_end = -1;
 
    while (PL_savestack_ix > save_start) {
-      const UV save_code = PL_savestack[PL_savestack_ix - 1].any_uv & SAVE_MASK;
+      const UV save_code = PL_savestack[PL_savestack_ix-1].any_uv & SAVE_MASK;
       int num_words = 0;
       // all kinds of save operations generated for localizing values
       switch (save_code) {
@@ -723,10 +753,10 @@ bool save_localizations(pTHX_ I32& start, I32& end)
          break;
       case SAVEt_DESTRUCTOR_X:
          num_words = 3;
-         if (PL_savestack_ix - 4 > save_start) {
-            const UV next_item = PL_savestack[PL_savestack_ix - 4].any_uv;
+         if (PL_savestack_ix-4 > save_start) {
+            const UV next_item = PL_savestack[PL_savestack_ix-4].any_uv;
             if ((next_item & SAVE_MASK) == SAVEt_ALLOC)
-               num_words += 1 + (next_item >> SAVE_TIGHT_SHIFT);
+               num_words += 1 + int(next_item >> SAVE_TIGHT_SHIFT);
          }
          break;
       }
@@ -736,12 +766,12 @@ bool save_localizations(pTHX_ I32& start, I32& end)
          if (tmp_save_end < 0 && last_kept != save_end) {
             EXTEND_MORTAL(save_end - save_start);
             PL_tmps_ix += save_end - save_start;
-            tmp_save_end = PL_tmps_ix;
+            tmp_save_end = I32(PL_tmps_ix);
             const I32 tmp_save_start = tmp_save_end - (save_end - last_kept);
             Copy(&(PL_savestack[last_kept]), &(PL_tmps_stack[tmp_save_start]), save_end - last_kept, ANY);
             last_kept = tmp_save_start;
          }
-         leave_scope(PL_savestack_ix - 1);
+         leave_scope(PL_savestack_ix-1);
       } else {
          PL_savestack_ix -= num_words;
          last_kept -= num_words;
@@ -769,7 +799,7 @@ int undo_saved_locals(pTHX_ SV* sv, MAGIC* mg)
 {
    const local_magic_t* lmg = static_cast<local_magic_t*>(mg);
    const I32 save_start = PL_savestack_ix;
-   const I32 num_saved = lmg->mg_len;
+   const I32 num_saved = I32(lmg->mg_len);
    SSGROW(num_saved);
    Copy(&(lmg->locals[0]), &(PL_savestack[save_start]), num_saved, ANY);
    PL_savestack_ix += num_saved;
@@ -792,7 +822,7 @@ OP* leave_local_block_op(pTHX)
       } else if (SvOK(sv) || SvTYPE(sv) != SVt_PVMG) {
          DIE(aTHX_ "local with: wrong storage value");
       }
-      const size_t mgsz = sizeof(local_magic_t) + (num_saved - 1) * sizeof(ANY);
+      const size_t mgsz = sizeof(local_magic_t)+(num_saved-1)*sizeof(ANY);
       char* mg_raw;
       Newxz(mg_raw, mgsz, char);
       local_magic_t* lmg = reinterpret_cast<local_magic_t*>(mg_raw);
@@ -813,20 +843,11 @@ OP* leave_local_block_op(pTHX)
 
 int parse_local_block(pTHX_ OP** op_ptr)
 {
-   lex_read_space(0);
-   if (PL_parser->bufptr == PL_parser->bufend || *PL_parser->bufptr != '(') {
+   op_keeper<OP> scope(aTHX_ parse_expression_in_parens(aTHX));
+   if (!scope) {
       report_parse_error("expected: local with(EXPR) { BLOCK }");
       return KEYWORD_PLUGIN_DECLINE;
    }
-   lex_read_to(PL_parser->bufptr + 1);
-   op_keeper<OP> scope(aTHX_ parse_termexpr(0));
-   if (!scope) return KEYWORD_PLUGIN_DECLINE;
-   lex_read_space(0);
-   if (PL_parser->bufptr == PL_parser->bufend || *PL_parser->bufptr != ')') {
-      report_parse_error("expected: local with(EXPR) { BLOCK }");
-      return KEYWORD_PLUGIN_DECLINE;
-   }
-   lex_read_to(PL_parser->bufptr + 1);
    lex_read_space(0);
    if (PL_parser->bufptr == PL_parser->bufend || *PL_parser->bufptr != '{') {
       report_parse_error("expected: local with(EXPR) { BLOCK }");
@@ -913,6 +934,71 @@ int parse_local_caller(pTHX_ OP** op_ptr)
    return KEYWORD_PLUGIN_EXPR;
 }
 
+// --------------------
+
+struct local_saveio_handler {
+   GV* gv;
+   GV* saved;
+
+   local_saveio_handler(pTHX_ GV* gv_)
+      : gv((GV*)SvREFCNT_inc_simple_NN(gv_))
+   {
+      if (GvIOp(gv)) {
+         saved = (GV*)newSV(0);
+         gv_init(saved, nullptr, "__ANONIO__", 10, 0);
+         if (do_openn(saved, ">&=", 3, FALSE, 0, 0, nullptr, (SV**)&gv, 1)) {
+            do_close(gv, FALSE);
+         } else {
+            SvREFCNT_dec(saved);
+            saved = nullptr;
+         }
+      } else {
+         saved = nullptr;
+      }
+   }
+
+   void undo (pTHX) const
+   {
+      if (GvIOp(gv)) do_close(gv, FALSE);
+      if (saved) {
+         (void)do_openn(gv, ">&=", 3, FALSE, 0, 0, nullptr, (SV**)&saved, 1);
+         SvREFCNT_dec(saved);
+      }
+      SvREFCNT_dec(gv);
+   }
+};
+
+OP* local_close_op(pTHX)
+{
+   dSP;
+   dPOPss;
+   if (SvTYPE(sv) != SVt_PVGV)
+      DIE(aTHX_ "not an IO handle in local close");
+   local_do<local_saveio_handler>(aTHX_ (GV*)sv);
+   RETURN;
+}
+
+OP* local_open_op(pTHX)
+{
+   SV* sv = PL_stack_base[TOPMARK+1];
+   if (SvTYPE(sv) != SVt_PVGV)
+      DIE(aTHX_ "not an IO handle in local open");
+   local_do<local_saveio_handler>(aTHX_ (GV*)sv);
+   return def_pp_OPEN(aTHX);
+}
+
+int parse_local_open_close(pTHX_ OP** op_ptr, OP* (*ppaddr)(pTHX))
+{
+   OP* o = parse_termexpr(0);
+   if (!o) return KEYWORD_PLUGIN_DECLINE;
+   o->op_ppaddr = ppaddr;
+   *op_ptr = o;
+   PL_hints |= HINT_BLOCK_SCOPE;
+   return KEYWORD_PLUGIN_EXPR;
+}
+
+// --------------------
+
 bool following_keyword(pTHX_ const AnyString& kw, bool skip_it = false)
 {
    if (PL_parser->bufptr + kw.len < PL_parser->bufend
@@ -925,6 +1011,22 @@ bool following_keyword(pTHX_ const AnyString& kw, bool skip_it = false)
    return false;
 }
 
+#if defined(POLYMAKE_GATHER_CODE_COVERAGE)
+int prevent_unnecessary_scope(pTHX_ int (*parse_func)(pTHX_ OP**), OP** op_ptr)
+{
+   const auto noopt = PERLDB_NOOPT;
+   PL_perldb &= ~PERLDBf_NOOPT;
+   const int result = parse_func(aTHX_ op_ptr);
+   PL_perldb |= noopt;
+   return result;
+}
+#else
+int prevent_unnecessary_scope(pTHX_ int (*parse_func)(pTHX_ OP**), OP** op_ptr)
+{
+   return parse_func(aTHX_ op_ptr);
+}
+#endif
+
 }
 
 int parse_enhanced_local(pTHX_ OP** op_ptr)
@@ -940,10 +1042,18 @@ int parse_enhanced_local(pTHX_ OP** op_ptr)
    case 'c':
       if (following_keyword(aTHX_ "caller", true))
          return parse_local_caller(aTHX_ op_ptr);
+      if (following_keyword(aTHX_ "close"))
+         return parse_local_open_close(aTHX_ op_ptr, local_close_op);
       break;
    case 'i':
       if (following_keyword(aTHX_ "if"))
-         return parse_local_if(aTHX_ op_ptr);
+         return prevent_unnecessary_scope(aTHX_ &parse_local_if, op_ptr);
+      if (following_keyword(aTHX_ "interrupts", true))
+         return parse_interrupts_op(aTHX_ true, op_ptr);
+      break;
+   case 'o':
+      if (following_keyword(aTHX_ "open"))
+         return parse_local_open_close(aTHX_ op_ptr, local_open_op);
       break;
    case 'p':
       if (following_keyword(aTHX_ "pop"))
@@ -971,13 +1081,22 @@ int parse_enhanced_local(pTHX_ OP** op_ptr)
       break;
    case 'w':
       if (following_keyword(aTHX_ "with", true))
-         return parse_local_block(aTHX_ op_ptr);
+         return prevent_unnecessary_scope(aTHX_ &parse_local_block, op_ptr);
       break;
    }
    return KEYWORD_PLUGIN_DECLINE;
 }
 
-} }
+}
+namespace ops {
+
+void init_globals(pTHX)
+{
+   def_pp_LEAVE = PL_ppaddr[OP_LEAVE];
+   def_pp_OPEN  = PL_ppaddr[OP_OPEN];
+}
+
+} } }
 
 // Local Variables:
 // mode:C++
